import chex
import jax.numpy as jnp
from typing import Optional, Callable
from .basis_measures import hippo_legs, hippo_legt, hippo_lagt, hippo_fout, HippoParams
from .diagonalization import scaled_diagonal, block_diagonal, diagonalize, eigenvector_transform


@chex.dataclass
class Hippo:
    """
    Base class for initializing HiPPOs (High-order Polynomial Projection Operators) derived from various basis measures. Uses the Chex dataclass rather than the standard library one for compatibility with JAX pytrees and therefore does not take positional arguments when being called.
    """

    state_size: int
    basis_measure: str
    conj_sym: bool = False
    dplr: bool = False
    diagonalize: bool = True
    block_diagonal: bool = False
    n_blocks: Optional[int] = None

    def __call__(self) -> HippoParams:
        if self.block_diagonal:
            assert self.n_blocks is not None
            self.state_size = self.state_size // self.n_blocks

        if self.basis_measure == 'legs':
            params = hippo_legs(self.state_size)

        elif self.basis_measure == 'legt':
            params = hippo_legt(
                self.state_size, lmu=False
            )

        elif self.basis_measure == 'lmu':
            params = hippo_legt(
                self.state_size, lmu=True,
            )

        elif self.basis_measure == 'lagt':
            params = hippo_lagt(
                self.state_size, generalized=False
            )

        elif self.basis_measure == 'glagt':
            params = hippo_lagt(
                self.state_size, generalized=True
            )

        elif self.basis_measure == 'fout':
            params = hippo_fout(
                self.state_size, diagonal=False,
                decay=False, double=False
            )

        elif self.basis_measure == 'foud':
            params = hippo_fout(
                self.state_size, diagonal=True,
                decay=False, double=False
            )

        elif self.basis_measure == 'fourier_decay':
            params = hippo_fout(
                self.state_size, diagonal=False,
                decay=True, double=False
            )

        elif self.basis_measure == 'fourier_double':
            params = hippo_fout(
                self.state_size, diagonal=False,
                decay=False, double=True
            )

        elif self.basis_measure == 'linear':
            params = scaled_diagonal(
                self.state_size, 'linear',
                self.conj_sym
            )

        elif self.basis_measure == 'inverse':
            params = scaled_diagonal(
                self.state_size, 'inverse',
                self.conj_sym
            )
        else:
            raise ValueError(
                'Invalid basis_measure argument:', self.basis_measure
            )

        if self.diagonalize:
            params = diagonalize(
                params, self.state_size, self.basis_measure,
                self.conj_sym, self.dplr
            )
            if self.block_diagonal:
                params = block_diagonal(params, self.state_size, self.n_blocks, self.conj_sym, self.basis_measure)

        self._params = params
        return self._params

    def b_initializer(self):
        """
        Input matrix parameter initializer for use in Modules across various jax neural net libraries i.e. Haiku, Flax etc...

        :return initializer function of signature: f(key, shape) if dplr is True, otherwise jnp.ones
        """
        if self.dplr:
            return lambda key, shape: self._params.input_matrix
        else:
            return jnp.ones

    def low_rank_initializer(self):
        """
        Low rank term parameter initializer for use in Modules across various jax neural net libraries i.e. Haiku, Flax etc...

        :return initializer function of signature: f(key, shape)

        """

        if self.dplr:
            return lambda key, shape: self._params.low_rank_term
        else:
            raise ValueError('Keep_input_matrix_and_low_term must be True')

    def lambda_initializer(self, return_array: str) -> Callable:
        """
        Splits state matrix back into real and imaginary parts for parameter initialization in Modules across various neural net libraries i.e. Haiku, Flax etc...

        :param return_array: str summoning either 'real', 'imaginary' or 'both'
        :return: initializer function of signature: f(key, shape)

        """
        if return_array == 'real':
            return lambda key, shape: self._params.state_matrix.real
        elif return_array == 'imaginary':
            return lambda key, shape: self._params.state_matrix.imag
        elif return_array == 'full':
            return lambda key, shape: self._params.state_matrix
        else:
            raise ValueError('Invalid return array:', return_array)

    def eigenvector_transform(
            self,
            input,
            inverse: bool = True,
            concatenate: bool = False
    ) -> Callable:
        """
        Wrapper method for global function of the same name which applies either a transformation of the input array by the inverse eigenvector or a transformation of the eigenvector by an input array. Unlike the function, this method assumes that the input array you are providing is the return value of a jax initialization function requiring key and shape arguments, such as jax.nn.initializers.normal(). It also provides an option for splitting the transformed array into real and imaginary parts and concatenating them back together for parameter initialization such as in the original S5 implementation: https://github.com/lindermanlab/S5

        :param input: input array
        :param inverse: bool deciding whether to apply inverse transformation
        :param concatenate: bool deciding whether to split and concatenate transformed array
        :return: transformed array

        """
        transformed = eigenvector_transform(self._params.eigenvector_pair, input, inverse, self.dplr)
        if concatenate:
            transformed_real = transformed.real
            transformed_imag = transformed.imag
            transformed = jnp.concatenate((transformed_real[..., None], transformed_imag[..., None]), axis=-1)
        return lambda key, shape: transformed



import jax.numpy as jnp
from jax.numpy import newaxis
from jax.numpy.linalg import eigh
from jax.scipy.linalg import block_diag
from typing import Tuple
from .basis_measures import HippoParams


def diagonalize(
        params: HippoParams,
        N: int,
        measure: str,
        conj_sym: bool,
        dplr: bool,

) -> HippoParams:
    """
    Decomposes the hippo state matrix into diagonal form, and diagonal plus low rank if dplr is True, which essentially means just keeping the input matrix and low rank term as well as the diagonal state matrix and it's eigenvectors.

    :param params: HippoParams object containing original state and input matrices
    :param N: state size
    :param measure: basis measure family
    :param conj_sym: bool deciding whether to enforce conjugate symmetry
    :param dplr:
    :return: HippoParams object containing diagonalized state matrix and it's eigenvectors

    """
    if measure in ['linear', 'inverse']:
        return params

    else:
        low_rank_term = low_rank(N, measure)
        if measure == 'legs':
            normal_plus_low_rank = params.state_matrix + low_rank_term[:, newaxis] * low_rank_term[newaxis, :]
        else:
            normal_plus_low_rank = params.state_matrix + jnp.sum(
                low_rank_term[:, newaxis] * low_rank_term[..., newaxis],
                axis=-3
            )

        Lambda_imag, eigenvector = eigh(normal_plus_low_rank * -1j)
        diagonal = jnp.diagonal(normal_plus_low_rank)
        Lambda_real = jnp.mean(diagonal) * jnp.ones_like(diagonal)
        Lambda = Lambda_real + 1j * Lambda_imag

        if conj_sym:
            Lambda = Lambda[:N // 2]
            eigenvector = eigenvector[:, :N // 2]

        inverse_eigenvector = eigenvector.conj().T
        eigenvector_pair = (eigenvector, inverse_eigenvector)

        if dplr:
            input_matrix = eigenvector_transform(eigenvector_pair, params.input_matrix, inverse=True, dplr=True)
            low_rank_term = eigenvector_transform(eigenvector_pair, low_rank_term, inverse=True, dplr=True)
            return HippoParams(
                state_matrix=Lambda,
                eigenvector_pair=eigenvector_pair,
                input_matrix=input_matrix,
                low_rank_term=low_rank_term
            )
        else:
            return HippoParams(
                state_matrix=Lambda,
                eigenvector_pair=eigenvector_pair,
            )


def scaled_diagonal(N: int, scaling: str, conj_sym: bool):
    """
    Creates either the linear or inverse approximations of the diagonal Hippo state matrix, the inverse being derived from the scaled Legendre (HiPPO-LegS) and the linear from the translated Fourier (HiPPO-FouT).

    :param N: state_size
    :param scaling: str specifying which scaling method to apply
    :param conj_sym: bool deciding whether to enforce conjugate symmetry
    :return: HippoParams object containing scaled_diagonal approximation as state matrix
    """
    if conj_sym:
        _N = N//2
    else:
        _N = N
    n = jnp.arange(_N)

    lambda_real = -0.5 * jnp.ones(_N)
    if scaling == 'linear':
        lambda_imaginary = jnp.pi * n
    elif scaling == 'inverse':
        lambda_imaginary = 1/jnp.pi * N * (N / (2*n + 1) - 1)
    else:
        raise ValueError(
            f'Not a valid scaling: {scaling}, options are confined to linear and inverse'
        )
    _lambda = lambda_real + 1j * lambda_imaginary

    return HippoParams(state_matrix=_lambda)


def block_diagonal(
        params: HippoParams,
        block_size: int,
        n_blocks: int,
        conj_sym: bool,
        measure: str,
) -> HippoParams:
    """
    Splits the diagonal Hippo state matrix and it's eigenvectors into blocks as in: https://arxiv.org/abs/2208.04933.

    :param params: HippoParams object containing state matrix and eigenvector pair
    :param block_size: size of each block
    :param n_blocks: number of blocks to split into
    :param conj_sym: bool deciding whether to enforce conjugate symmetry
    :param measure: string specifying the measure family of input HippoParams
    :return: HippoParams object containing state matrix and eigenvector pair in block structure

    """
    if conj_sym:
        block_size = block_size // 2

    state_matrix = (params.state_matrix * jnp.ones((n_blocks, block_size))).ravel()
    if measure in ['linear', 'inverse']:
        return HippoParams(state_matrix=state_matrix)
    else:
        eigenvector = block_diag(*([params.eigenvector_pair[0]] * n_blocks))
        inverse_eigenvector = block_diag(*([params.eigenvector_pair[1]] * n_blocks))
        return HippoParams(state_matrix=state_matrix, eigenvector_pair=(eigenvector, inverse_eigenvector))


def low_rank(
        N: int,
        measure: str,
) -> jnp.ndarray:
    """
    Creates the low rank term for various measure families.

    :param N: state size
    :param measure: basis measure family
    :return: jnp.ndarray
    """
    n = jnp.arange(N)

    if measure == 'legs':
        P = jnp.sqrt(n + 0.5)

    elif measure == 'legt':
        P = jnp.sqrt(2 * n + 1)
        P0 = P.at[0::2].set(0)
        P1 = P.at[1::2].set(0)
        P = jnp.stack([P0, P1], axis=0)
        P *= 2 ** (-0.5)

    elif measure == 'lagt':
        P = jnp.sqrt(0.5) * jnp.ones((1, N))

    elif measure == 'fout':
        P = jnp.zeros(N)
        P = P.at[0::2].set(jnp.sqrt(2))
        P = P.at[0].set(1)
        P = P[newaxis, :]

    elif measure == 'fourier_decay':
        P = jnp.zeros(N)
        P = P.at[0::2].set(jnp.sqrt(2))
        P = P.at[0].set(1)
        P = P[newaxis, :]
        P = P / jnp.sqrt(2)

    elif measure == 'fourier_double':
        P = jnp.zeros(N)
        P = P.at[0::2].set(jnp.sqrt(2))
        P = P.at[0].set(1)
        P = P[newaxis, :]
        P = jnp.sqrt(2) * P

    elif measure == 'foud':
        P = jnp.zeros((1, N))

    else:
        raise ValueError(
            'invalid measure argument:', measure
        )

    return P


def eigenvector_transform(
        eigenvector_pair: Tuple[jnp.ndarray, jnp.ndarray],
        array: jnp.ndarray, inverse: bool, dplr: bool
) -> jnp.ndarray:
    """
    Applies either a transformation of the input array by the inverse eigenvector or a transformation of the eigenvector by an input array. See wrapper method of base Hippo class for an example in a parameter initialization context.

    :param eigenvector_pair: tuple containing the eigenvectors indexed from regular to inverse
    :param array: input n-dimensional array
    :param inverse: decides whether to apply inverse transformation
    :param dplr: bool specifying whether the input is in diagonal plus low rank representation
    :return: transformed array

    """
    if inverse:
        if dplr:
            return jnp.einsum('ij, ...j -> ...i', eigenvector_pair[1], array)
        else:
            return eigenvector_pair[1] @ array
    else:
        array = array[..., 0] + 1j * array[..., 1]
        return array @ eigenvector_pair[0]





import jax.numpy as jnp
import jax.scipy.special as jss
import scipy.special as ss
from jax.numpy import newaxis
from typing import Tuple, NamedTuple, Optional


class HippoParams(NamedTuple):
    """
    Base class for storing matrices directly derived from HiPPO.
    """

    state_matrix: jnp.ndarray
    eigenvector_pair: Optional[Tuple[jnp.ndarray, jnp.ndarray]] = None
    input_matrix: Optional[jnp.ndarray] = None
    low_rank_term: Optional[jnp.ndarray] = None


def hippo_legs(N: int) -> HippoParams:
    """
    Creates the scaled Legendre basis measure HiPPO

    :param N: state size
    :return: HippoParams object containing state and input matrices

    """
    n = jnp.arange(N)
    polynomial = jnp.sqrt(2 * n + 1)

    A = polynomial[:, newaxis] * polynomial[newaxis, :]
    A = jnp.tril(A) - jnp.diag(n)

    A = -A
    B = polynomial

    return HippoParams(state_matrix=A, input_matrix=B)


def hippo_legt(
        N: int,
        lmu: bool,
) -> HippoParams:
    """
    Creates the translated Legendre basis measure HiPPO

    :param N: state_size
    :param lmu: bool deciding whether to use Legendre memory unit initialization used in: https://proceedings.neurips.cc/paper/2019/file/952285b9b7e7a1be5aa7849f32ffff05-Paper.pdf
    :return: HippoParams object containing state and input matrices

    """
    n = jnp.arange(N)
    j, i = jnp.meshgrid(n, n)
    polynomial = (2 * n + 1)

    if lmu:
        A = jnp.where(i < j,
                      x=-1,
                      y=(-1.) ** (i - j + 1),
                      ) * (polynomial[:, newaxis])

        B = (-1.) ** n[:, newaxis] * (polynomial[:, newaxis])

    else:
        polynomial = jnp.sqrt(polynomial)

        A = jnp.where(i < j,
                      x=(-1.) ** (i - j),
                      y=1,
                      )
        A = polynomial[:, newaxis] * A * polynomial[newaxis, :]

        A = -A
        B = polynomial[: newaxis]

        A *= 0.5
        B *= 0.5

    return HippoParams(state_matrix=A, input_matrix=B)


def hippo_lagt(
        N: int,
        generalized: bool,
        beta: float = 1.0,
) -> HippoParams:
    """
    Creates the translated Laguerre basis measure HiPPO

    :param N: state size
    :param generalized: bool deciding whether to use the generalized initialization (GLagT)
    :param beta: constant
    :return: HippoParams object containing state and input matrices

    """
    if generalized:
        n = jnp.arange(N)
        alpha = 0.0
        beta = 0.01

        A = -jnp.eye(N) * (1 + beta) \
            / 2 - (jnp.tril(jnp.ones((N, N)), -1))

        B = ss.binom(alpha + n, n)[:, newaxis]

        polynomial = jnp.exp(0.5 * (jss.gammaln(n + alpha + 1) - jss.gammaln(n + 1)))

        A = (1. / polynomial[:, newaxis]) * A * polynomial[newaxis, :]
        B = (1. / polynomial[:, newaxis]) * B * (
                jnp.exp(-0.5 * jss.gammaln(1 - alpha)) * beta ** ((1 - alpha) / 2)
        )

    else:
        A = jnp.eye(N) \
            / 2 - (jnp.tril(jnp.ones((N, N))))

        B = beta * jnp.ones(N)

    return HippoParams(state_matrix=A, input_matrix=B)


def hippo_fout(
        N: int,
        diagonal: bool,
        decay: bool,
        double: bool,
) -> HippoParams:
    """
    Creates the translated Fourier basis measure HiPPO

    :param N: state size
    :param diagonal: bool deciding whether to use the diagonal initialization (FouD)
    :param decay: bool deciding whether to use the decay initialization (Fourier_decay)
    :param double: bool deciding whether to use the double initialization (Fourier_double)
    :return: HippoParams object containing state and input matrices

    """
    if N % 2 != 0:
        raise ValueError('N must be an even value for this projection')
    frequencies = jnp.arange(N // 2)
    if double:
        frequencies = frequencies * 2

    if diagonal:
        stack = jnp.stack(
            [frequencies, jnp.zeros(N // 2)],
            axis=-1
        ).reshape(-1)[:-1]

        A = 2 * jnp.pi * (-jnp.diag(stack, 1) + jnp.diag(stack, -1))

    else:
        stack = jnp.stack(
            [jnp.zeros(N // 2), frequencies],
            axis=-1
        ).reshape(-1)[1:]

        A = jnp.pi * (-jnp.diag(stack, 1) + jnp.diag(stack, -1))

    B = jnp.zeros(N)
    B = B.at[0::2].set(jnp.sqrt(2))
    B = B.at[0].set(1)

    if diagonal:
        A = A - 0.5 * jnp.eye(N)
    elif decay:
        A = A - 0.5 * B[:, newaxis] * B[newaxis, :]
        B = 0.5 * B
    elif double:
        A = A - B[:, newaxis] * B[newaxis, :] * 2
        B = B * 2
    else:
        A = A - B[:, newaxis] * B[newaxis, :]

    return HippoParams(state_matrix=A, input_matrix=B)
